/* C compiler: statement parsing */

#include "c.h"
int isinsyncmode = 0;   /* global flag to pass mode of syncness to function calls CWK*/
int isinsideastartbody = 0; /* global flag to know whether a start occured*/
static int SEQcounter = 0;  /*counts SEQ statements, used to generate labels*/

#define SWITCHSIZE 512      /* initial size of switch arrays */

float density = 0.5;      /* minimum switch density */
int refinc = 1000;      /* amount by which ref field is incremented */

static struct equate {   /* deferred equates: */
   Symbol new;      /* new label, i.e., label being defined */
   Symbol old;      /* old label */
   struct equate *oldlink;   /* pointer to equate structure for old=x */
   struct equate *link;   /* next equate on list */
} *equlist;         /* list of deferred equates */
struct swtch {      /* switch statement data: */
   Symbol sym;      /* local holding the switch value */
   int lab;      /* label of selection code */
   Symbol deflab;      /* default label */
   int ncases;      /* number of cases */
   int size;      /* size of value & label arrays */
   int    *values;      /* value, label pairs */
   Symbol *labels;
};

dclproto(static void branch,(int));
dclproto(static void caselabel,(struct swtch *, int, int));
dclproto(static void cmp,(int, Symbol, int, int));
dclproto(static void dostmt,(int, struct swtch *, int, int)); /*CWK*/
dclproto(static void xxxdostmt,(int, struct swtch *, int, int)); /*CWK*/
dclproto(static int foldcond,(Tree, Tree));
dclproto(static void forstmt,(int, struct swtch *, int, int));
dclproto(static void xxxforstmt,(int, struct swtch *, int, int)); /*CWK*/
dclproto(static void forkstmt,(int, struct swtch *, int, int, int, int));   /*CWK*/
dclproto(static void startstmt,(int, struct swtch *, int));   /*CWK*/
dclproto(static void ifstmt,(int, int, struct swtch *, int, int, int, int));
dclproto(static void xxxifstmt,(int, int, struct swtch *, int, int, int, int)); /*CWK*/
dclproto(static Symbol localaddr,(Tree));
dclproto(static void stmtlabel,(char *));
dclproto(static int swcode,(struct swtch *, int *, int, int, int));
dclproto(static void swgen,(struct swtch *));
dclproto(static void swstmt,(int, int, int, int, int, int)); /*CWK*/
dclproto(static void visit,(struct equate *));
dclproto(static void whilestmt,(int, struct swtch *, int, int)); /*CWK*/
dclproto(static void xxxwhilestmt,(int, struct swtch *, int, int)); /*CWK*/

extern int may_dep_on_pr_var( Tree );   /*CWK, in tree.c*/


/* branch - jump to lab */
static void branch(lab) {
   Code cp;

   walk(0, 0, 0);
   for (cp = codelist; cp->kind < Label; )
      cp = cp->prev;
   if (lab)
      while (cp->kind == Label && islabel(cp->u.node)
      && cp->u.node->syms[0]->u.l.label != lab) {
         Symbol old = findlabel(lab);
         equatelab(cp->u.node->syms[0], old);
         old->ref++;
         if (cp->prev->next = cp->next)
            cp->next->prev = cp->prev;
         else
            codelist = cp->prev;
         do
            cp = cp->prev;
         while (cp->kind < Label);
      }
   if (cp->kind != Jump) {
      code(Jump);
      codelist->u.node = lab ? jump(lab) : 0;
   }
   /* else do nothing and thus delete dead branch */
}

/* caselabel - add a label to the current switch list */
static void caselabel(swp, val, lab) struct swtch *swp; {
   int k;

   if (swp->ncases >= swp->size) {
      int    *vals = swp->values;
      Symbol *labs = swp->labels;
      swp->size *= 2;
      swp->values = (int    *) talloc(swp->size*sizeof (int));
      swp->labels = (Symbol *) talloc(swp->size*sizeof (Symbol));
      for (k = 0; k < swp->ncases; k++) {
         swp->values[k] = vals[k];
         swp->labels[k] = labs[k];
      }
   }   
   for (k = swp->ncases; k > 0 && swp->values[k-1] >= val; k--) {
      swp->values[k] = swp->values[k-1];
      swp->labels[k] = swp->labels[k-1];
   }
   if (k < swp->ncases && swp->values[k] == val)
      error("duplicate case label `%d'\n", val);
   swp->values[k] = val;
   swp->labels[k] = findlabel(lab);
   ++swp->ncases;
   if (Aflag >= 2 && swp->ncases == 258)
      warning("more than 257 cases in a switch statement\n");
}

/* cmp - generate code for `if (p op n) goto lab' for integer n */
static void cmp(op, p, n, lab) Symbol p; {
   listnodes(eqnode(op, cast(idnode(p), inttype), constnode(n, inttype)), lab, 0);
}

/* definelab - define a label */
void definelab(lab) {
   Code cp;

   walk(0, 0, 0);
   code(Label);
   codelist->u.node = newnode(LABEL+V, 0, 0, findlabel(lab));
   for (cp = codelist; cp->kind <= Label; )
      cp = cp->prev;
   while (cp->kind == Jump && cp->u.node && cp->u.node->kids[0]
   && cp->u.node->kids[0]->op == ADDRG+P
   && cp->u.node->kids[0]->syms[0]->u.l.label == lab) {
      cp->u.node->kids[0]->syms[0]->ref--;
      assert(cp->next);
      cp->prev->next = cp->next;
      cp->next->prev = cp->prev;
      cp = cp->prev;
   }
}

/* definept - define an execution point n: current token at current pc */
Code definept(p) Coordinate *p; {
   int n;
   Code cp = code(Defpoint);

   cp->u.point.src = p ? *p : src;
   cp->u.point.point = npoints;
   if (ncalled > 0 && (n = findcount(src.file, src.x, src.y)) >= 0)
      refinc = 1000*n/ncalled + 1;
   if (events.points) {
      Tree e = 0;
      apply(events.points, (Generic)&cp->u.point.src, (Generic)&e);
      if (e)
         listnodes(e, 0, 0);
   }
   if (glevel > 2)
      locus(identifiers, &cp->u.point.src);
   return cp;
}

/* dostmt - do statement while ( expression ) */
static void dostmt(lab, swp, lev, rframes) struct swtch *swp; {
   refinc *= 10;
   t = gettok();
   definelab(lab);
   statement(lab, swp, lev, 0, 0, rframes);        /*CWK*/
   definelab(lab + 1);
   expect(WHILE);
   expect('(');
   definept(0);
   walk(conditional(')'), lab, 0);
   if (findlabel(lab + 2)->ref)
      definelab(lab + 2);
}

/* xxxdostmt - xxxdo statement while ( expression ) */
static void xxxdostmt(lab, swp, lev, rframes) struct swtch *swp; {
   Tree e;  /*CWK*/
   extern void enlgrp(int), size(), exlgrp(), synchronize();  /*CWK*/
   int priv;   /*CWK*/
   /* CWK:  Da das statement zuerst geemitted wird, aber erst nach Parsen von
    * e bekannt ist, ob private Pcode-Befehle eingefuegt werden muessen,
    * ist hier eine Erzeugung von Do-Code fuer private Abbruchbedingungen 
    * nur mit vielen Labels und einer temporaeren Variable moeglich:
    * L:   statement
    * L+1: e                          continue-Ansprungadresse
    *      if false, jump to L+4;     ggf. aufraeumen und weiter
    * L+2: if tmp jump to L+3;        break-Ansprungadresse
    *      tmp = 1;                   tmp=1, falls ein enlgrp/size angelegt wurde
    *      size;                      Die allererste Iteration wird immer,
    *      enlgrp;                     also mit shared Bedingung 1 ausgefuehrt.
    *      jump to L;
    * L+3: k nops;                    f"ur alle weiteren Iterationen,
    *      jump to L;                  wenn enlgrp/size schon angelegt
    * L+4: if !tmp jump to L+5;       exlgrp ausfuehren g.d.w. enlgrp angelegt
    *      exlgrp;
    * L+5: .....
    * Wegen der Umstaendlichkeit dieser Implementierung wird die Realisierung
    * eines privaten do-stmt mit 4 labels ohne temp. Var. wie folgt vereinfacht:
    * L+3: enlgrp
    * L:   size
    *      statement
    * L+1: e
    *      if true, jump to L    (L+1 = continue-Ansprungadresse)
    * L+2: exlgrp; sync; ....    (L+2 = break-Ansprungadresse)
    * d.h., die Bedingung wird IMMER als privat angenommen
    * (konservative worst-case-Annahme).
    * Bis auf size() muesste das fuer shared loopconditions 
    * sowieso gemacht werden, wegen potentieller breaks.  */
   refinc *= 10;
   t = gettok();
   definelab(lab+3);  /* Dummy, um das Emitten der bisherigen Stmts zu erzwingen*/
   enlgrp( 2 );
   definelab(lab);
   size();
   statement(lab, swp, lev, 0, 0, rframes + 1);
   definelab(lab + 1);
   expect(WHILE);
   expect('(');
   definept(0);
   e = conditional(')');
   walk(e, lab, 0);
   if (findlabel(lab + 2)->ref)
      definelab(lab + 2);
   exlgrp();
   synchronize();  /*CWK*/
}

/* equatelab - add new=old to list of deferred equates */
void equatelab(new, old) Symbol new, old; {
   struct equate *e = (struct equate *) talloc(sizeof *e);

   e->new = new;
   e->old = old;
   e->oldlink = 0;
   e->link = equlist;
   equlist = e;
   for (e = e->link; e; e = e->link)
      if (e->old == new)
         e->oldlink = equlist;
      else if (e->new == old)
         equlist->oldlink = e;
}
   
/* flushequ - flush deferred equates */
void flushequ() {
   for ( ; equlist; equlist = equlist->link)
      visit(equlist);
}

/* foldcond - check if initial test in for(e1;e2;e3) S is necessary */
static int foldcond(e1, e2) Tree e1, e2; {
   Opcode op = generic(e2->op);
   Symbol v;

   if (e1 == 0 || e2 == 0)
      return 0;
   if (generic(e1->op) == ASGN && isaddrop(e1->kids[0]->op)
   && generic(e1->kids[1]->op) == CNST) {
      v = e1->kids[0]->u.sym;
      e1 = e1->kids[1];
   } else
      return 0;
   if ((op==LE || op==LT || op==EQ || op==NE || op==GT || op==GE)
   && generic(e2->kids[0]->op) == INDIR
   && e2->kids[0]->kids[0]->u.sym == v
   && e2->kids[1]->op == e1->op) {
      e1 = simplify(op, e2->type, e1, e2->kids[1]);
      if (e1->op == CNST+I)
         return e1->u.v.i;
   }
   return 0;
}

/* forstmt - for ( [expr1] ; [expr2] ; [expr3] ) statement */
static void forstmt(lab, swp, lev, rframes) struct swtch *swp; {
   Tree e1, e2, e3;
   Coordinate pt2, pt3;
   static char follow[] = { IF, ID, '}', 0 };

   e1 = e2 = e3 = 0;
   t = gettok();
   expect('(');
   definept(0);
   if (kind[t] == ID)
      e1 = expr0(';');
   else
      expect(';');
   walk(e1, 0, 0);
   pt2 = src;
   refinc *= 10;
   if (kind[t] == ID)
      e2 = texpr(conditional, ';');
   else
      expect(';');
   pt3 = src;
   if (kind[t] == ID)
      e3 = texpr(expr0, ')');
   else
      test(')', follow);
   if (glevel == 2) {   /* e1; L: if (!e2) goto L+2; S; L+1: e3; goto L; L+2: */
      definelab(lab);
      definept(&pt2);
      if (e2)
         walk(e2, 0, lab + 2);
      statement(lab, swp, lev, 0, 0, rframes);  /*CWK*/
      definelab(lab + 1);
      definept(&pt3);
      if (e3)
         walk(e3, 0, 0);
      branch(lab);
   } else {   /* e1; goto L+3; L: S; L+1: e3; L+3: if (e2) goto L; L+2: */
      int once = 0;
      if (e2) {
         once = foldcond(e1, e2);
         if (!once)
            branch(lab + 3);
      }
      definelab(lab);
      statement(lab, swp, lev, 0, 0, rframes);  /*CWK*/
      definelab(lab + 1);
      definept(&pt3);
      if (e3)
         walk(e3, 0, 0);
      if (e2) {
         if (!once)
            definelab(lab + 3);
         definept(&pt2);
         walk(e2, lab, 0);
      } else {
         definept(&pt2);
         branch(lab);
      }
   }
   if (findlabel(lab + 2)->ref)
      definelab(lab + 2);
}

/* xxxforstmt - xxxfor ( [expr1] ; [expr2] ; [expr3] ) statement */
static void xxxforstmt(lab, swp, lev, rframes) struct swtch *swp; {
   Tree e1, e2, e3;
   Coordinate pt2, pt3;
   static char follow[] = { IF, ID, '}', 0 };
   int priv, reallypriv;     /*CWK*/
   extern void enlgrp(int), exlgrp(), synchronize(), size();

   e1 = e2 = e3 = 0;
   t = gettok();
   expect('(');
   definept(0);
   if (kind[t] == ID)
      e1 = expr0(';');
   else
      expect(';');
   walk(e1, 0, 0);
   pt2 = src;
   refinc *= 10;
   if (kind[t] == ID)
      e2 = texpr(conditional, ';');
   else
      expect(';');
   if (e2) priv = (may_dep_on_pr_var( e2 )); /*CWK*/
   else    priv = 0;  /*CWK*/
   reallypriv = priv;  /*CWK, spare size ein fuer shared loop condition */
   priv = 1;   /*CWK: wegen breaks muss immer private uebersetzt werden!*/
   pt3 = src;
   if (kind[t] == ID)
      e3 = texpr(expr0, ')');
   else
      test(')', follow);
   if (glevel == 2) {   /* e1; L: if (!e2) goto L+2; S; L+1: e3; goto L; L+2: */
      if (priv) enlgrp( 1 );    /*CWK*/
      definelab(lab);
      definept(&pt2);
      if (e2)
         walk(e2, 0, lab + 2);
      if (priv && reallypriv) size();    /*CWK, vorher priv */
      statement(lab, swp, lev, 0, 0, rframes + ((priv)? 1:0) ); /*CWK*/
      definelab(lab + 1);
      definept(&pt3);
      if (e3)
         walk(e3, 0, 0);
      branch(lab);
      /* keine nops if (priv);   weil am Ende synchronisiert wird. CWK*/
   } else {   /* e1; goto L+3; L: S; L+1: e3; L+3: if (e2) goto L; L+2: */
      int once = 0;
      if (priv) enlgrp( 1 );    /*CWK*/
      if (e2) {
         once = foldcond(e1, e2);
         if (!once)
            branch(lab + 3);
      }
      definelab(lab);
      if (priv) size();        /*CWK*/
      statement(lab, swp, lev, 0, 0, rframes + ((priv)?1:0));  /*CWK*/
      definelab(lab + 1);
      definept(&pt3);
      if (e3)
         walk(e3, 0, 0);
      if (e2) {
         if (!once)
            definelab(lab + 3);
         definept(&pt2);
         walk(e2, lab, 0);
      } else {
         definept(&pt2);
         branch(lab);
      }
      /* keine nops, weil am Ende synchronisiert wird. CWK*/
   }
   if (findlabel(lab + 2)->ref)
      definelab(lab + 2);
   if (priv) exlgrp();    /*CWK*/
   /*if (priv)*/ synchronize();   /*CWK*/
}

/* forkstmt - fork ( [expr1] ; [expr2] ; [expr3] ) statement */    /*CWK*/
static void forkstmt(lab, swp, lev, bframes, cframes, rframes) struct swtch *swp; {
   Tree e1, e2, e3;
   Coordinate pt2, pt3;
   static char follow[] = { IF, ID, '}', 0 };
   extern void mkgrp(), engrp(), exgrp(), synchronize(), ibp(int);  /* in gen.c */

   e1 = e2 = e3 = 0;
   t = gettok();
   expect('(');
   definept(0);
   if (kind[t] == ID) {
      e1 = expr(';');
      if (may_dep_on_pr_var( e1 ))  /*CWK*/
         warning("fork: number of groups must not depend on private value!\n");
   }
   else {
      error("fork: number of groups required!\n");
      expect(';');
   }
   if (e1) {
      e1 = tree(MKGRPV, voidtype, e1, 0 );    /*CWK*/
      code(Gen); 
      codelist->u.node = listnodes(e1, 0, 0);
   }  /* spoolout */
   definept(0);
   refinc *= 10;
   if (kind[t] == ID) {
      e2 = expr0(';');
      /*
      if (generic(e2->op) == ASGN && 
          (generic(e2->kids[1]->op)==DIV || generic(e2->kids[1]->op)==MOD) )
      e2 = tree(ASGN, e2->type, e2->kids[0],
             simplify( e2->kids[1]->op, e2->kids[1]->type, e2->kids[1]->kids[0],
                                                           e2->kids[1]->kids[1]) );
      */
   }
   else {
      error("fork statement: new group number expression for @ expected!\n");
      expect(';');
   }
   walk(e2, 0, 0);
   definept(0);
   engrp();
   if (kind[t] == ID) {
      e3 = texpr(expr0, ')');
      walk(e3, 0, 0);
   }
   else if (t==')') {
      warning("fork: no expression renaming $!\n");
      t = gettok();
   }
   else {
      error("fork: expression renaming $ expected!\n");
      test(')',follow);
   }
   definept(0);
   if (glevel) ibp( 6 );      /*950504*/
   /* definelab(lab);  break/continue aus der fork-Anweisung moeglich 17.9. */
   statement(lab, swp, lev, bframes + 1, cframes + 1, rframes + 1);    /*CWK*/
   exgrp();
   synchronize();    /* nach Ruecksprache mit Helmut Seidl */
}

/* syncstmt - start ( expression ) statement */                /*CWK*/
static void syncstmt(lab, swp, lev) struct swtch *swp; {
   Tree e1;

   t = gettok();
   e1 = tree(SYNCV, voidtype, 0, 0 );    /*CWK*/
   code(Gen); 
   codelist->u.node = listnodes(e1, 0, 0); /* spoolout */
}

static void startstmt(lab, swp, lev) struct swtch *swp; {
   Tree e;
   extern void sproc(), xproc();
   extern int shframedepth; 
   extern int shoffset;
   int oldshoffset;	

   t = gettok();
   if (t=='(')  {  /*Expression follows*/
     error("Old-fashioned `start(k) {...}' is no longer supported,\n\
                     use  `start if ($<k) {...}' or join() instead!\n");
     expect('(');
     definept(0);
     e = expr(')');
     if (!isint(e->type)) {
        error("illegal type `%t' in start expression\n", e->type);
        e = retype(e, inttype);
     }
     /*e = cast(e, promote(e->type));*/
     if (e) {
        if (may_dep_on_pr_var(e))
           warning("number of started processes must be shared\n");
        e = tree( SPROCV, voidtype, e, 0 );
        code(Gen); 
        codelist->u.node = listnodes(e, 0, 0);
     }  /* enforce spoolout*/
     definelab(lab);              /*damit kein break aus dem start rausfuehren kann*/
   }
   else {   /*default: no expression -> start all available processors*/
     e = tree( SPROCV, voidtype, 0, 0 );
     code(Gen); 
     codelist->u.node = listnodes(e, 0, 0);
   }
   isinsyncmode = 1; isinsideastartbody = 1;
   shframedepth ++;  
   /*testweise raus 970908 oldshoffset = shoffset; shoffset = 0;   /*970903*/
   statement(lab, swp, lev, 0, 0, 0);   /*CWK: no former frames to flush. */
   shframedepth --;  
   /* testweise raus 970908 shoffset = oldshoffset;   /*970903*/
   isinsyncmode = 0; isinsideastartbody = 0;
   xproc();
}


/* ifstmt - if ( expression ) statement [ else statement ] */
static void ifstmt(lab, loop, swp, lev, bframes, cframes, rframes) struct swtch *swp; {/*CWK*/
   t = gettok();
   expect('(');
   definept(0);
   if ((refinc /= 2) == 0)
      refinc = 1;
   walk(conditional(')'), 0, lab);
   statement(loop, swp, lev, bframes, cframes, rframes);/*CWK*/
   if (t == ELSE) {
      branch(lab + 1);
      t = gettok();
      definelab(lab);
      statement(loop, swp, lev, bframes, cframes, rframes);/*CWK*/
      if (findlabel(lab + 1)->ref)
         definelab(lab + 1);
   } else
      definelab(lab);
}

/* xxxifstmt - xxxif ( expression ) statement [ else statement ] */
static void xxxifstmt(lab, loop, swp, lev, bframes, cframes, rframes) struct swtch *swp; {/*CWK*/
   Tree e;       /*CWK*/
   int priv;     /*CWK*/
   extern void splitSM();  /*CWK, alle in gen.c */
   extern void nextcase();   /*CWK*/
   extern void lgrp_if();   /*CWK*/
   extern void synchronize();   /*CWK*/
   t = gettok();
   expect('(');
   definept(0);
   e = conditional(')');
   priv = (may_dep_on_pr_var( e )); /*CWK*/
   if ((refinc /= 2) == 0)
      refinc = 1;
   if (priv) splitSM();      /*CWK: freien Speicher gleichmaessig aufteilen */
   walk(e, 0, lab);
   if (priv) nextcase(0);    /*CWK: then-Gruppe aufmachen */
   statement(loop, swp, lev, bframes+((priv)?1:0), cframes+((priv)?1:0), rframes + ((priv)?1:0));/*CWK*/
   if (t == ELSE) {
      branch(lab + 1);    /* zum lgrp_if springen */
      t = gettok();
      definelab(lab);
      if (priv) nextcase(1); /*CWK: else-Gruppe aufmachen */
      statement(loop, swp, lev, bframes+((priv)?1:0), cframes+((priv)?1:0), rframes + ((priv)?1:0));/*CWK*/
      if (findlabel(lab + 1)->ref)
         definelab(lab + 1);
   } else {  /*dummy-Else, um die idle Prozessoren zu beschaeftigen: */
      if (priv) branch(lab + 1);    /* zum lgrp_if springen */
      definelab(lab);
      if (priv) nextcase(1); /*CWK: else-Gruppe aufmachen */
      if (priv && findlabel(lab + 1)->ref)
         definelab(lab + 1);
   }
   if (priv) lgrp_if();  /*CWK: beide Gruppen zumachen */
   if (priv) synchronize();  /*CWK: synchr. im alten Frame */
}

/* localaddr - returns q if p yields the address of local/parameter q;
   otherwise returns 0 */
static Symbol localaddr(p) Tree p; {
   if (p == 0)
      return 0;
   switch (generic(p->op)) {
   case INDIR: case CALL: case ARG: case RARG: case SARG:  /*CWK*/
      return 0;
   case ADDRL: case ADDRF:
   case ADDRSL: case ADDRSF:   /*neu 970902 CWK*/
      return p->u.sym;
   case RIGHT: case ASGN:
      if (p->kids[1])
         return localaddr(p->kids[1]);
      return localaddr(p->kids[0]);
   case COND: {
      Symbol q;
      if (p->kids[1] && (q = localaddr(p->kids[1])))
         return q;
      return localaddr(p->kids[2]);
      }
   default: {
      Symbol q;
      if (p->kids[0] && (q = localaddr(p->kids[0])))
         return q;
      return localaddr(p->kids[1]);
      }
   }
}

/* retcode - return p from the current function */
void retcode(p, lab, rframes) Tree p; {
   extern int returnflush[50], returnstartflush[50], returnpos;
        /*CWK, Flags in gen.c */
   if (p == 0) {
      if (events.returns)
         apply(events.returns, (Generic)cfunc, (Generic)0);
      p = tree(RET+V, voidtype, 0, 0);
   } else {
      Type ty;
      p = pointer(p);
      if (ty = assign(freturn(cfunc->type), p))
         p = cast(p, ty);
      else
         error("illegal return type; found `%t' expected `%t'\n",
            p->type, freturn(cfunc->type));
      if (cfunc->shared) { /*CWK*/
         if (may_dep_on_pr_var( p ))
            warning("shared valued function may return a private value\n");
      }
      if (retv) {
         if (p->op == RIGHT && p->kids[0]->op == CALL+B
         && p->kids[1]->op == INDIR+B) {
            p = p->kids[0];
            p->kids[1] = idnode(retv);
         } else
            p = asgnnode(ASGN, rvalue(idnode(retv)), p);
         walk(p, 0, 0);
         if (events.returns)
            apply(events.returns, (Generic)cfunc, (Generic)rvalue(idnode(retv)));
         p = tree(RET+V, voidtype, 0, 0);
      } else {
         if (events.returns) {
            Symbol t1 = genident(AUTO, p->type, level);
            addlocal(t1);
            walk(asgn(t1, p), 0, 0);
            apply(events.returns, (Generic)cfunc, (Generic)idnode(t1));
            p = idnode(t1);
         }
         p = cast(p, promote(p->type));
         if (isptr(p->type)) {
            Symbol q = localaddr(p);
            if (q && (q->computed || q->generated))
               warning("pointer to a %s is an illegal return value\n",
                  q->scope == PARAM ? "parameter" : "local");
            else if (q)
               warning("pointer to %s `%s' is an illegal return value\n",
                  q->scope == PARAM ? "parameter" : "local", q->name);
            p = cast(p, unsignedtype);
         }
         if (cfunc->shared)    /*CWK*/
               p = tree(SRET + widen(p->type), p->type, p, 0);
         else  p = tree(RET + widen(p->type), p->type, p, 0);
      }
   }
   assert( rframes >= 0 );   /*CWK*/
   p->rframes = rframes;     /*CWK*/
   p->started = isinsideastartbody; /*!(cfunc->sync) && isinsyncmode;     /*CWK*/
   if (isinsyncmode && cfunc && !(cfunc->sync))              /*CWK*/
      warning("return crossing start/join boundary may destroy synchronity\n"); 
   walk(p, 0, 0);
   branch(lab);
}

/* statement - parse statements */
void statement(loop, swp, lev, bframes, cframes, rframes) struct swtch *swp; {  /*CWK*/
   extern void synchronize();   /*CWK*/
   int lab, ref = refinc;

   if (Aflag >= 2 && lev == 15)
      warning("more than 15 levels of nested statements\n");
   switch (t) {
   case IF:
      if (isFORKmodul && isinsyncmode)    /*CWK*/
        xxxifstmt(genlabel(2), loop, swp, lev + 1, bframes, cframes, rframes);/*CWK*/
      else
        ifstmt(genlabel(2), loop, swp, lev + 1, bframes, cframes, rframes);  /*CWK*/
      break;
   case WHILE:
      if (isFORKmodul && isinsyncmode)        /*CWK*/
        xxxwhilestmt(genlabel(3), swp, lev + 1, rframes );   /*CWK*/
      else 
        whilestmt(genlabel(3), swp, lev + 1, rframes); /*CWK*/
      break;
   case DO:
      if (isFORKmodul && isinsyncmode)
        xxxdostmt(genlabel(4), swp, lev + 1, rframes );   /*CWK*/
      else
        dostmt(genlabel(3), swp, lev + 1, rframes); /*CWK*/
      expect(';');
      break;
   case FOR:
      if (isFORKmodul && isinsyncmode)
         xxxforstmt(genlabel(4), swp, lev + 1, rframes);   /*CWK*/
      else
         forstmt(genlabel(4), swp, lev + 1, rframes); /*CWK*/
      break;
   case FORK:                                          /*CWK*/
      if (!isinsyncmode)
         error("fork statement is not allowed in asynchronous context\n");
      forkstmt(loop, swp, lev + 1, bframes, cframes, rframes );
      break;
   case START:                                          /*CWK*/
      if (isinsyncmode)
         error("start statement is not allowed in synchronous context\n");
      startstmt(genlabel(1), swp, lev + 1 );
      break;
   case SWITCH:
      swstmt(loop, genlabel(2), lev + 1, bframes, cframes, rframes);  /*CWK*/
      break;
   case CASE:
      if (swp == 0)
         error("illegal case label\n");
      definelab(lab = genlabel(1));
      while (t == CASE) {
         Tree p;
         t = gettok();
         p = constexpr(':');
         if (generic(p->op) == CNST && isint(p->type)) {
            if (swp) {
               needconst++;
               p = cast(p, swp->sym->type);
               needconst--;
               caselabel(swp, p->u.v.i, lab);
            }
         } else
            error("case label must be a constant integer expression\n");
         ntree = 0;
      }
      statement(loop, swp, lev, bframes, cframes, rframes);  /*CWK*/
      break;
   case DEFAULT:
      if (swp)
         if (swp->deflab)
            error("extra default label\n");
         else {
            swp->deflab = findlabel(swp->lab);
            definelab(swp->deflab->u.l.label);
         }
      else
         error("illegal default label\n");
      t = gettok();
      expect(':');
      statement(loop, swp, lev, bframes, cframes, rframes);
      break;
   case SEQ: {                        /*970310 CWK*/
      extern void begseq(int), endseq(int);
      int mylabelnr = SEQcounter++;
      int ignore = 0;
      if (!isFORKmodul) {
         ignore = 1;
         error("seq is a FORK-specific statement!\n");
      }
      if (!isinsyncmode) {
         ignore = 1;
         warning("seq in asynchronous context is ignored\n");
      }
      t = gettok();
      isinsyncmode = 0;
      walk(0,0,0);   /*damit vorangehende stmts geflusht werden 970612*/
      if (!ignore)  begseq( mylabelnr );
      statement(loop, swp, lev, bframes, cframes, rframes );
      if (! ignore) {
         endseq( mylabelnr );
         isinsyncmode = 1;
      }
      break;
      }
   case FARM: {                                                       /*CWK*/
      extern void ibp( int );
      int ignore = 0;
      if (!isFORKmodul) {
         ignore = 1;
         error("farm is a FORK-specific statement!\n");
      }
      if (!isinsyncmode) {
         ignore = 1;
         warning("farm in asynchronous context is ignored\n");
      }
      t = gettok();
      isinsyncmode = 0;
      if (glevel && !ignore)  ibp( 7 );         /*950504*/
      statement(loop, swp, lev, bframes, cframes, rframes );
      if (! ignore) {
         synchronize();     /* else ignore synchronization */
         if (glevel) ibp( 8 );      /*950504*/
         isinsyncmode = 1;
      }
      break;
      }
   case BEGINSYNC:        /* 950912 */
      if (isinsyncmode)
         error("beginsync statement is not allowed in synchronous context\n");
      t = gettok();
      isinsyncmode = 1; isinsideastartbody = 1;
      statement(loop, swp, lev, 0, 0, 0);
      isinsyncmode = 0; isinsideastartbody = 0;
      break;
   case RETURN:
      t = gettok();
      definept(0);
      if (cfunc->sync && !isinsyncmode)
         error("return from inside a farm statement is forbidden!\n");
      if (!isinsyncmode && isinsideastartbody)  /*970821*/
         warning("return crossing start/join and farm/seq boundary may destroy synchronity\n"); 
      if (t != ';')
         if (freturn(cfunc->type) == voidtype) {
            error("extraneous return value\n");
            expr(0);
            retcode(0, cfunc->u.f.label, rframes);
         } else
            retcode(expr(0), cfunc->u.f.label, rframes);
      else {
         if (freturn(cfunc->type) != voidtype
         && (freturn(cfunc->type) != inttype || Aflag >= 1))
            warning("missing return value\n");
         retcode(0, cfunc->u.f.label, rframes);
      }
      expect(';');
      break;
   case BREAK:
      walk(0, 0, 0);
      if (swp && swp->lab > loop) {
         definept(0);
         branch(swp->lab + 1);
         t = gettok();
         expect(';');
         break;
      }
   case CONTINUE:
      walk(0, 0, 0);
      definept(0);
      if (loop) {
         extern void flushframes( int );                  /*CWK*/
         flushframes( (t == BREAK)? bframes : cframes );  /*CWK*/
         branch(loop + (t == BREAK ? 2 : 1));
      }
      else
         error("illegal %k statement\n", t);
      t = gettok();
      expect(';');
      break;
   case '{':
      compound(loop, swp, lev + 1, bframes, cframes, rframes);  /*CWK*/
      break;
   case ';':
      definept(0);
      t = gettok();
      break;
   case GOTO:
      walk(0, 0, 0);
      definept(0);
      t = gettok();
      if (t == ID) {
         Symbol p = lookup(token, labels[0]);
         if (p == 0) {
            p = install(token, &labels[0], 0);
            p->u.l.label = genlabel(1);
            p->u.l.equatedto = p;
            p->src = src;
         }
         p->ref++;
         use(p, src);
         branch(p->u.l.label);
         t = gettok();
      } else
         error("missing label in goto\n");
      expect(';');
      break;
   case ID:
      if (getchr() == ':') {
         stmtlabel(token);
         statement(loop, swp, lev, bframes, cframes, rframes);
         break;
      }
   default:
      definept(0);
      if (kind[t] != ID) {
         error("unrecognized statement\n");
         t = gettok();
      } else {
         Tree e = expr0(0);
         listnodes(e, 0, 0);
         if (glevel > 1 || nodecount == 0 || nodecount > 200)
            walk(0, 0, 0);
         ntree = 0;
      }
      expect(';');
   }
   if (kind[t] != IF && kind[t] != ID && t != '}' && t != EOI) {
      static char follow[] = { IF, ID, '}', 0 };
      error("illegal statement termination\n");
      skipto(0, follow);
   }
   refinc = ref;
}

/* stmtlabel - label : */
static void stmtlabel(label) char *label; {
   Symbol p;

   if ((p = lookup(label, labels[0])) == 0) {
      p = install(label, &labels[0], 0);
      p->u.l.label = genlabel(1);
      p->u.l.equatedto = p;
   }
   if (p->defined)
      error("redefinition of label `%s' previously defined at %w\n",
         label, &p->src);
   p->defined = 1;
   p->src = src;
   definelab(p->u.l.label);
   t = gettok();
   expect(':');
}

/* swcode - generate switch decision code for buckets b[lb..ub] */
static int swcode(swp, b, lb, ub, n) struct swtch *swp; int b[]; {
   int hilab, k, l, lolab, median, u, *v;

   v = swp->values;
   median = ((float)v[b[lb]] + (float)v[b[ub+1]-1] + 1)/2;
   for (l = lb, u = ub; l <= u; ) {
      k = (l + u)/2;
      if (v[b[k]] > median)
         u = k - 1;
      else if (v[b[k+1]-1] < median)
         l = k + 1;
      else
         break;
   }
   if (l > u) {   /* which of two buckets is closest to the median? */
      l = b[k];
      u = b[k+1] - 1;
      if (median < v[l] && k > lb && v[l] - median > median - v[l-1])
         k--;
      else if (median > v[u] && k < ub && median - v[u] > v[u+1] - median)
         k++;
   }
   l = b[k];
   u = b[k+1] - 1;
   lolab = hilab = swp->deflab->u.l.label;
   if (k > lb) {
      lolab = 0;
      if (k < ub)
         hilab = genlabel(1);
   } else if (k < ub)
      hilab = 0;
   if (n == 0)
      lolab = hilab = 0;
   if (u - l + 1 <= 3) {
      int i;
      for (i = l; i <= u; i++)
         cmp(EQ, swp->sym, swp->values[i], swp->labels[i]->u.l.label);
      if (lolab && lolab == hilab)
         branch(lolab);
      else {
         if (lolab)
            cmp(LT, swp->sym, swp->values[l], lolab);
         if (hilab)
            cmp(GT, swp->sym, swp->values[u], hilab);
         walk(0, 0, 0);
      }
   } else {
      int lab = genlabel(1);
      Symbol tab = genident(STATIC, array(voidptype, u - l + 1, 0), GLOBAL);
      cmp(LT, swp->sym, swp->values[l], lolab ? lolab : lab);
      cmp(GT, swp->sym, swp->values[u], hilab ? hilab : lab);
      walk(tree(JUMP, voidtype, rvalue((*opnode['+'])(ADD, pointer(idnode(tab)),
         (*opnode['-'])(SUB, cast(idnode(swp->sym), inttype),
            constnode(swp->values[l], inttype)))), 0), 0, 0);
      code(Switch);
      codelist->u.swtch.sym = swp->sym;
      codelist->u.swtch.deflab = swp->deflab;
      codelist->u.swtch.table = tab;
      codelist->u.swtch.size = u - l + 1;
      codelist->u.swtch.values = &swp->values[l];
      codelist->u.swtch.labels = &swp->labels[l];
      branch(0);
      definelab(lab);
      if (swp->values[u] - swp->values[l] + 1 >= 10000)
         warning("switch generates a huge table\n");
   }
   if (k > lb)
      n = swcode(swp, b, lb, k - 1, n - 1);
   if (k < ub) {
      if (hilab)
         definelab(hilab);
      n = swcode(swp, b, k + 1, ub, n - 1);
   }
   return n;
}

/* swgen - partition case labels into buckets, initiate code generation */
#define den(k,n) ((float)(k-buckets[n]+1)/(float)(v[k]-v[buckets[n]]+1))
static void swgen(swp) struct swtch *swp; {
   int *buckets, k, n, *v;

   if (swp->ncases == 0)
      return;
   buckets = (int *) talloc((swp->ncases + 1)*sizeof *buckets);
   v = swp->values;
   n = 0;
   for (k = 0; k < swp->ncases; k++, n++) {
      buckets[n] = k;
      while (n > 0) {
         float d = den(k, n-1);
         if (d < density || k < swp->ncases - 1 && d < den(k+1, n))
            break;
         n--;
      }
   }
   buckets[n--] = swp->ncases;
   swcode(swp, buckets, 0, n, n);
}

/* swstmt - switch ( expression ) statement */
static void swstmt(loop, lab, lev, bframes, cframes, rframes) {  /*CWK*/
   Tree e;
   struct swtch sw;
   Code head, tail;

   t = gettok();
   expect('(');
   definept(0);
   e = expr(')');
   if (!isint(e->type)) {
      error("illegal type `%t' in switch expression\n", e->type);
      e = retype(e, inttype);
   }
   if (isinsyncmode && may_dep_on_pr_var( e ))  /*CWK 961112*/
      warning("private switch selector in synchronous mode will probably cause deadlock!\n");
   e = cast(e, promote(e->type));
   if (generic(e->op) == INDIR && isaddrop(e->kids[0]->op)
   && e->kids[0]->u.sym->type == e->type && !isvolatile(e->kids[0]->u.sym->type)) {
      sw.sym = e->kids[0]->u.sym;
      walk(0, 0, 0);
   } else {
      sw.sym = genident(REGISTER, e->type, level);
      addlocal(sw.sym);
      walk(asgn(sw.sym, e), 0, 0);
   }
   code(Switch);
   head = codelist;
   sw.lab = lab;
   sw.deflab = 0;
   sw.ncases = 0;
   sw.size = SWITCHSIZE;
   sw.values = (int    *) talloc(SWITCHSIZE*sizeof (int));
   sw.labels = (Symbol *) talloc(SWITCHSIZE*sizeof (Symbol));
   if ((refinc /= 10) == 0)
      refinc = 1;
   statement(loop, &sw, lev, 0, cframes, rframes);  /*CWK*/
   /* Bei Privatisierung muss hier         ^^^^ +1 hin. CWK */
   if (sw.deflab == 0) {
      sw.deflab = findlabel(lab);
      definelab(lab);
      if (sw.ncases == 0)
         warning("switch statement with no cases\n");
   }
   if (findlabel(lab + 1)->ref)
      definelab(lab + 1);
   tail = codelist;
   codelist = head->prev;
   swgen(&sw);
   branch(lab);
   head->next->prev = codelist;
   codelist->next = head->next;
   codelist = tail;
}

/* visit - equate p->new=p->old, equating predecessor first */
static void visit(p) struct equate *p; {
   if (p && p->old) {
      visit(p->oldlink);
      p->new->u.l.equatedto = p->old->u.l.equatedto;
      p->old = 0;
   }
}

/* whilestmt - while ( expression ) statement */
static void whilestmt(lab, swp, lev, rframes) struct swtch *swp; {
   Coordinate pt;
   Tree e;

   refinc *= 10;
   t = gettok();
   expect('(');
   walk(0, 0, 0);
   pt = src;
   e = texpr(conditional, ')');
   if (generic(e->op) == CNST || glevel == 2) {
         /* L, L+1: if (!e) goto L+2; S; goto L; L+2: */
      definelab(lab);
      definelab(lab + 1);
      definept(&pt);
      walk(e, 0, lab + 2);
      statement(lab, swp, lev, 0, 0, rframes);   /*CWK*/
      branch(lab);
   } else {   /* goto L+1; L: S; L+1: if (E) goto L; L+2: */
      branch(lab + 1);
      definelab(lab);
      statement(lab, swp, lev, 0, 0, rframes);  /*CWK*/
      definelab(lab + 1);
      definept(&pt);
      walk(e, lab, 0);
   }
   if (findlabel(lab + 2)->ref)
      definelab(lab + 2);
}

/* xxxwhilestmt - xxxwhile ( expression ) statement */
static void xxxwhilestmt(lab, swp, lev, rframes) struct swtch *swp; {
   Coordinate pt;
   Tree e;
   int priv;        /*CWK*/
   extern void enlgrp(int), exlgrp(), size(), synchronize();

   refinc *= 10;
   t = gettok();
   expect('(');
   walk(0, 0, 0);
   pt = src;
   e = texpr(conditional, ')');
   priv = (may_dep_on_pr_var( e )); /*CWK*/
   priv = 1;   /*CWK: wegen break notwendig!*/
   if (generic(e->op) == CNST || glevel == 2) {   /*CWK: CNST = shared!*/
         /* L, L+1: if (!e) goto L+2; S; goto L; L+2: */
      definelab(lab);
      definelab(lab + 1);
      definept(&pt);
      walk(e, 0, lab + 2);
      statement(lab, swp, lev, 0, 0, rframes + ((priv)?1:0)); /*CWK*/
      branch(lab);
   } else {   /* goto L+1; L: S; L+1: if (E) goto L; L+2: */
      if (priv) enlgrp( 0 );    /*CWK*/
      branch(lab + 1);
      definelab(lab);
      if (priv) size();     /*CWK*/
      statement(lab, swp, lev, 0, 0, rframes + ((priv)?1:0)); /*CWK*/
      definelab(lab + 1);
      definept(&pt);
      walk(e, lab, 0);
   }
   if (findlabel(lab + 2)->ref)
      definelab(lab + 2);
   /* keine Nops, weil am Ende synchronisiert wird.  CWK*/
   if (priv) exlgrp();    /*CWK*/
   synchronize();
   /*CWK, es muss auch bei shared Bedingung immer synchronisiert werden*/
}
